/*
Copyright 2023 by Valentina Gliozzi, Gian Luca Pozzato, Gabriele Tessore, Alberto Valese

    This file is part of CLean.

    CLean is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CLean is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CLean.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Theorem prover CLean */
:- op(400,fy,neg), 
   op(500,xfy,and), 
   op(600,xfy,or),
   op(650,xfy,->), 
   op(650,xfy,=>).
:-use_module(library(lists)).


/* --------------------------------------------------------------------------------------- */

/* The following predicate implements the calculi:

   prove(Gamma,Delta,Labels,RCond,LCond,CEM,MP,ProofTree) :- Gamma |- Delta 
   
 */

pr(Fml):-prove([[],[],[]],[[],[],[[x,Fml]]],[x],[],[],[],[],[],_).
pr(Fml,Tree):-prove([[],[],[]],[[],[],[[x,Fml]]],[x],[],[],[],[],[],Tree).

/* Axioms */
/* true and false */
prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],_,_,_,_,_,_,tree(falso,null,null,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
  member([_,false],LitGamma),!.
prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],_,_,_,_,_,_,tree(vero,null,null,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
  member([_,true],LitDelta),!.
prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],_,_,_,_,_,_,tree(assioma,null,null,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
  member(F,LitGamma),member(F,LitDelta),!.
prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],_,_,_,_,_,_,tree(assioma,null,null,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
  member(F,TransGamma),member(F,TransDelta),!.
prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],_,_,_,_,_,_,tree(assioma,null,null,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
  member(F,ComplexGamma),member(F,ComplexDelta),!.

  /* Rules */
  /* and L */
  prove([LitGamma,TransGamma,ComplexGamma],Delta,Labels,RCond,LCond,
        CEM,MP,CS,tree(andL,SubTree,null,[LitGamma,TransGamma,ComplexGamma],Delta)):-
    select([X,A and B],ComplexGamma,ResComplexGamma),!,
    put([X,A],LitGamma,ResComplexGamma,TempLitGamma,TempComplexGamma),
    put([X,B],TempLitGamma,TempComplexGamma,NewLitGamma,NewComplexGamma),
    prove([NewLitGamma,TransGamma,NewComplexGamma],Delta,Labels,
        RCond,LCond,CEM,MP,CS,SubTree).
  /* or R */
  prove(Gamma,[LitDelta,TransDelta,ComplexDelta],Labels,RCond,LCond,
        CEM,MP,CS,tree(orR,SubTree,null,Gamma,[LitDelta,TransDelta,ComplexDelta])):-
    select([X,A or B],ComplexDelta,ResComplexDelta),!,
    put([X,A],LitDelta,ResComplexDelta,TempLitDelta,TempComplexDelta),
    put([X,B],TempLitDelta,TempComplexDelta,NewLitDelta,NewComplexDelta),
    prove(Gamma,[NewLitDelta,TransDelta,NewComplexDelta],Labels,
        RCond,LCond,CEM,MP,CS,SubTree).
  /* neg R */
  prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,tree(negR,SubTree,null,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
    select([X,neg A],ComplexDelta,ResComplexDelta),!,
    put([X,A],LitGamma,ComplexGamma,NewLitGamma,NewComplexGamma),
    prove([NewLitGamma,TransGamma,NewComplexGamma],[LitDelta,TransDelta,ResComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,SubTree).
  /* neg L */
  prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,tree(negL,SubTree,null,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
    select([X,neg A],ComplexGamma,ResComplexGamma),!,
    put([X,A],LitDelta,ComplexDelta,NewLitDelta,NewComplexDelta),
    prove([LitGamma,TransGamma,ResComplexGamma],[NewLitDelta,TransDelta,NewComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,SubTree).
  /* -> R */
  prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,tree(impR,SubTree,null,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
    select([X,A -> B],ComplexDelta,ResComplexDelta),!,
    put([X,A],LitGamma,ComplexGamma,NewLitGamma,NewComplexGamma),
    put([X,B],LitDelta,ResComplexDelta,NewLitDelta,NewComplexDelta),
    prove([NewLitGamma,TransGamma,NewComplexGamma],[NewLitDelta,TransDelta,NewComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,SubTree).
  /* CS: additional parameter for avoiding multiple applications */
  prove(Gamma,[LitDelta,TransDelta,ComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,tree(cs,LeftProof,RightProof,Gamma,[LitDelta,TransDelta,ComplexDelta])):-
    member([X,A => B],ComplexDelta),
    \+member([X,A => B],CS),!,
    put([X,A],LitDelta,ComplexDelta,LitDelta1,ComplexDelta1),
    put([X,B],LitDelta,ComplexDelta,LitDelta2,ComplexDelta2),
    prove(Gamma,[LitDelta1,TransDelta,ComplexDelta1],
      Labels,RCond,LCond,CEM,MP,[[X,A => B]|CS],LeftProof),
    prove(Gamma,[LitDelta2,TransDelta,ComplexDelta2],
      Labels,RCond,LCond,CEM,MP,[[X,A => B]|CS],RightProof).
  /* => R */
  prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],
        Labels,RCond,LCond,CEM,MP,CS,tree(condR,SubTree,null,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
    member([X,A => B],ComplexDelta),
    \+member([X,A => B],RCond),!,
    genFreshLabel(Y,Labels),
    put([Y,B],LitDelta,ComplexDelta,NewLitDelta,NewComplexDelta),
    prove([LitGamma,[[X,A,Y]|TransGamma],ComplexGamma],[NewLitDelta,TransDelta,NewComplexDelta],
      [Y|Labels],[[X,A => B]|RCond],LCond,CEM,MP,CS,SubTree).
  /* and R */
  prove(Gamma,[LitDelta,TransDelta,ComplexDelta],Labels,RCond,LCond,
        CEM,MP,CS,tree(andR,LeftProof,RightProof,Gamma,[LitDelta,TransDelta,ComplexDelta])):-
    select([X,A and B],ComplexDelta,ResComplexDelta),!,
    put([X,A],LitDelta,ResComplexDelta,FirstLitDelta,FirstComplexDelta),
    put([X,B],LitDelta,ResComplexDelta,SecLitDelta,SecComplexDelta),
    prove(Gamma,[FirstLitDelta,TransDelta,FirstComplexDelta],
        Labels,RCond,LCond,CEM,MP,CS,LeftProof),
    prove(Gamma,[SecLitDelta,TransDelta,SecComplexDelta],
        Labels,RCond,LCond,CEM,MP,CS,RightProof).
  /* or L */
  prove([LitGamma,TransGamma,ComplexGamma],Delta,Labels,
      RCond,LCond,CEM,MP,CS,tree(orL,LeftProof,RightProof,[LitGamma,TransGamma,ComplexGamma],Delta)):-
    select([X,A or B],ComplexGamma,ResComplexGamma),!,
    put([X,A],LitGamma,ResComplexGamma,FirstLitGamma,FirstComplexGamma),
    put([X,B],LitGamma,ResComplexGamma,SecLitGamma,SecComplexGamma),
    prove([FirstLitGamma,TransGamma,FirstComplexGamma],Delta,
        Labels,RCond,LCond,CEM,MP,CS,LeftProof),
    prove([SecLitGamma,TransGamma,SecComplexGamma],Delta,
        Labels,RCond,LCond,CEM,MP,CS,RightProof).
  /* -> L */
  prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,tree(impL,LeftTail,RightTail,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
    select([X,A -> B],ComplexGamma,ResComplexGamma),!,
    put([X,A],LitDelta,ComplexDelta,NewLitDelta,NewComplexDelta),
    put([X,B],LitGamma,ResComplexGamma,NewLitGamma,NewComplexGamma),
    prove([LitGamma,TransGamma,ResComplexGamma],[NewLitDelta,TransDelta,NewComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,LeftTail),
    prove([NewLitGamma,TransGamma,NewComplexGamma],[LitDelta,TransDelta,ComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,RightTail).
   /* MP: additional parameter for avoiding multiple applications */
  prove(Gamma,[LitDelta,TransDelta,ComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,tree(mp,SubTree,null,Gamma,[LitDelta,TransDelta,ComplexDelta])):-
    member([X,A,X],TransDelta),
    \+member([X,A,X],MP),!,
    put([X,A],LitDelta,ComplexDelta,NewLitDelta,NewComplexDelta),
    prove(Gamma,[NewLitDelta,TransDelta,NewComplexDelta],
      Labels,RCond,LCond,CEM,CS,[[X,A,X]|MP],SubTree).  
   /* => L  */
  prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,tree(condL,LeftTail,RightTail,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
    member([X,A => B],ComplexGamma),
    member(Y,Labels),
    \+member([Y,[X,A => B]],LCond),!,
    prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,[[X,A,Y]|TransDelta],ComplexDelta],
      Labels,RCond,[[Y,[X,A => B]]|LCond],CEM,MP,CS,LeftTail),
    put([Y,B],LitGamma,ComplexGamma,ResLitGamma,ResComplexGamma),
    prove([ResLitGamma,TransGamma,ResComplexGamma],[LitDelta,TransDelta,ComplexDelta],
      Labels,RCond,[[Y,[X,A => B]]|LCond],CEM,MP,CS,RightTail).
   /* CEM: additional parameter for avoiding multiple applications */
  prove(Gamma,[LitDelta,TransDelta,ComplexDelta],
      Labels,RCond,LCond,CEM,MP,CS,tree(cem,LeftProof,RightProof,Gamma,[LitDelta,TransDelta,ComplexDelta])):-
    member([X,A => B],ComplexDelta),
    \+member([X,A => B],CEM),!,
    member(Y,Labels),
    prove(Gamma,[LitDelta,[[X,A,Y]|TransDelta],ComplexDelta],
      Labels,RCond,LCond,[[X,A => B]|CEM],MP,CS,LeftProof),
    put([Y,B],LitDelta,ComplexDelta,NewLitDelta,NewComplexDelta),
    prove(Gamma,[NewLitDelta,TransDelta,NewComplexDelta],
      Labels,RCond,LCond,[[X,A => B]|CEM],MP,CS,RightProof).
  /* EQ */
  prove([LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta],_,_,_,_,_,_,tree(eq,LeftProof,RightProof,[LitGamma,TransGamma,ComplexGamma],[LitDelta,TransDelta,ComplexDelta])):-
    member([X,A,Y],TransGamma),member([X,B,Y],TransDelta),
    /* The premises are "independent" from the conclusion, so we can use label x
       for their proofs */
    put([x,A],[],[],FirstLitGamma,FirstComplexGamma),
    put([x,B],[],[],FirstLitDelta,FirstComplexDelta),
    put([x,B],[],[],SecLitGamma,SecComplexGamma),
    put([x,A],[],[],SecLitDelta,SecComplexDelta),
    prove([FirstLitGamma,[],FirstComplexGamma],[FirstLitDelta,[],FirstComplexDelta],
      [x],[],[],[],[],[],LeftProof),
    prove([SecLitGamma,[],SecComplexGamma],[SecLitDelta,[],SecComplexDelta],
      [x],[],[],[],[],[],RightProof).
  
  
  /* This predicate generates a "new" label for that branch */
  genFreshLabel(x,Labels):-true,\+member(x,Labels),!.
  genFreshLabel(y,Labels):-true,\+member(y,Labels),!.
  genFreshLabel(z,Labels):-true,\+member(z,Labels),!.
  genFreshLabel(u,Labels):-true,\+member(z,Labels),!.
  genFreshLabel(v,Labels):-true,\+member(z,Labels),!.
  genFreshLabel(w,Labels):-true,\+member(z,Labels),!.
  genFreshLabel(k,Labels):-true,\+member(k,Labels),!.
  genFreshLabel(j,Labels):-true,\+member(j,Labels),!.
  genFreshLabel(i,Labels):-true,\+member(i,Labels),!.
  genFreshLabel(x+1,[i|_]):-!.
  genFreshLabel(x+N1,[Last|_]):-Last=x+N,N1 is N+1,!.
  
  /* This predicate inserts a formula in the correct list; more precisely, a formula
   can be an atomic formula or a complex formula */
put([X,A and B],Literal,Formulae,Literal,[[X,A and B]|Formulae]):-!.
put([X,A or B],Literal,Formulae,Literal,[[X,A or B]|Formulae]):-!.
put([X,A -> B],Literal,Formulae,Literal,[[X,A -> B]|Formulae]):-!.
put([X,A => B],Literal,Formulae,Literal,[[X,A => B]|Formulae]):-!.
put([X,neg A],Literal,Formulae,Literal,[[X,neg A]|Formulae]):-!.
put(X,Literal,Formulae,[X|Literal],Formulae):-!.

