import ast,re
from clean_string import clean_string
class Sequent:
    def __init__(self):
        self.lit = []
        self.trans = []
        self.complex = []


    def __str__(self, depth=0):
        lit_str = "[" + ", ".join(["[" + ", ".join(sublit) + "]" for sublit in self.lit]) + "]"
        trans_str = "[" + ", ".join(["[" + ", ".join(subtrans) + "]" for subtrans in self.trans]) + "]"
        complex_str = "[" + ", ".join([str(complex_tree) for complex_tree in self.complex]) + "]"
        tabs = "\t" * depth
        return f"{tabs}#\tLit: {lit_str}\n{tabs}#\tTrans: {trans_str}\n{tabs}#\tComplex: {complex_str}"

    def to_latex(self, depth=0):
        #return "sequent"
        tabs = "\t" * depth
        lit_str = ", ".join([f"{lit[0]}: {lit[1].upper()}" for lit in self.lit])
        trans_str = ", ".join([" ".join([f"{trans[i]} \\trans{{{trans[i+1].upper()}}} {trans[i+2]}" for i in range(0, len(trans), 3)]) for trans in self.trans])
        complex_str = ", ".join(["{} : {}".format(complex_tree[0], self.replace_latex(complex_tree[1].upper())) for complex_tree in self.complex])
        return ", ".join(filter(None, [lit_str, trans_str, complex_str]))

    def replace_latex(self,s):
        #print(s)
        x = {'=>':' \\cond ', '->':' \\imp ', 'AND':'\\andd', 'OR':'\\orr', 'NEG':'\\nott'}
        for k,v in x.items():
            s = s.replace(k,v)
        return s

class Tree:
    def __init__(self):
        self.motivation = ""
        self.left_proof = None
        self.right_proof = None
        self.gamma = Sequent()
        self.delta = Sequent()
    
    def __str__(self):
        left_proof_str = str(self.left_proof) if self.left_proof else "None"
        right_proof_str = str(self.right_proof) if self.right_proof else "None"
        return f"\nTree:\n#\tMotivation: {self.motivation}\n#\tLeft Proof: {left_proof_str}\n#\tRight Proof: {right_proof_str}\n#\tGamma:\n{str(self.gamma)}\n#\tDelta:\n{str(self.delta)}"

    def __repr__(self, depth=0):
        left_proof_str = self.left_proof.__repr__(depth + 1) if self.left_proof else "None"
        right_proof_str = self.right_proof.__repr__(depth + 1) if self.right_proof else "None"
        tabs = "\t" * depth
        return f"\n{tabs}Tree:\n{tabs}#\tMotivation: {self.motivation}\n{tabs}#\tLeft Proof: {left_proof_str}\n{tabs}#\tRight Proof: {right_proof_str}\n{tabs}#\tGamma:\n{self.gamma.__str__(depth + 1)}\n{tabs}#\tDelta:\n{self.delta.__str__(depth + 1)}"

    def to_latex(self, depth=0):
        tabs = "\t" * depth
        gamma_latex = self.gamma.to_latex()
        delta_latex = self.delta.to_latex()
        justifies = f"\\justifies {gamma_latex} \\prova {delta_latex} \\using ({{\\bf {self.mot_to_latex(self.motivation.upper())}}})"
        left_proof_str = self.left_proof.to_latex(depth + 1) if self.left_proof else ""
        right_proof_str = self.right_proof.to_latex(depth + 1) if self.right_proof else ""
        return f"\\[\n{left_proof_str}\n{right_proof_str}\n{justifies}\\]" if any([left_proof_str, right_proof_str, justifies]) else ""

    def mot_to_latex(self,s):
        d = {'ASSIOMA':'AX', 'CONDL':'\\cond {\\bf L}','CONDR':'\\cond {\\bf R}','IMPR':'\\imp {\\bf R}','IMPL':'\\imp {\\bf L}','ANDL':'\\andd {\\bf L}',
            'ANDR':'\\andd {\\bf R}', 'ORR':'\\orr {\\bf R}', 'ORL':'\\orr {\\bf L}', 'NEGL':'\\nott {\\bf L}', 'NEGR':'\\nott {\\bf R}'}
        for k,v in d.items():
            s = s.replace(k,v)
        return s

#def clean_string(s):
    #return s.replace("Atom('437509')",'y').replace("Atom('273413')",'a').replace("Atom('428933')",'x').replace("Functor(8708365,2,a,a)",'a=>a')

def string_to_list(input_string):
    input_string
    try:
        # Utilizza il modulo ast per valutare la stringa come espressione Python
        #print(input_string)
        def wrap_non_strings(match):
            return f"'{match.group(0)}'"

        # Adding quotes around non-string elements
        input_string = re.sub(r'(?<=\[|\,)\s*([^,\[\]]+)(?=\s|\]|,|$)', wrap_non_strings, input_string)
        #print(input_string)
        result = ast.literal_eval(input_string)
        return result
    except (SyntaxError, ValueError) as e:
        print(f"Errore nella conversione della stringa: {input_string} {e}")
        return None

def parse_sequent(s, sequent_obj):
    s = s.strip()
    s = s[1:-1]  # Rimuovi le parentesi quadre iniziali e finali
    token = ''
    elements = []
    depth = 0
    for i in range(len(s)):
        token += s[i]
        if s[i] == '[':
            depth +=1
        elif s[i] == ']':
            depth -=1

        if s[i] == ' ' and s[i-1] == ']' and depth == 0:
            elements.append(token[:-1])
            token = ''
        elif s[i] == ' ' and s[i-1] == ']' and depth != 0:
            token += ','
            

    elements.append(token)
    #print(elements)
    #print(string_to_list(elements[0]))
    if len(elements) == 3:
        sequent_obj.lit = string_to_list(elements[0])
        sequent_obj.trans = string_to_list(elements[1])
        
        #print(elements)
        sequent_obj.complex = string_to_list(elements[2])

def parse_tree(s):
    s = s.strip()
    if s.startswith("tree("):
        s = s[5:-1]  # Rimuovi il prefisso "tree(" e la parentesi finale ")"
        
        # Trova tutti i token principali all'interno del nodo tree
        main_tokens = []
        token = ""
        depth = 0
        in_brackets = False
        for char in s:
            if char == ',' and depth == 0 and not in_brackets:
                main_tokens.append(token)
                token = ""
            else:
                token += char
                if char == '(':
                    depth += 1
                elif char == ')':
                    depth -= 1
                if char == '[':
                    in_brackets = True
                elif char == ']':
                    in_brackets = False
        main_tokens.append(token)  # Aggiungi l'ultimo token
        
        # abbiamo i main, (tree e motivazione), adesso prendiamo i sequent


        s = ''.join(main_tokens[3:])

        sequents = []
        token = ''
        for char in s:
            token += char
            if char == '[':
                depth += 1
            elif char == ']':
                depth -= 1
                if  depth == 0:
                    sequents.append(token)
                    token = ""

        #print(sequents)

        tree_obj = Tree()
        tree_obj.motivation = main_tokens[0]
        
        left_proof_str = main_tokens[1] if len(main_tokens) > 1 else "null"
        if left_proof_str != "null":
            tree_obj.left_proof = parse_tree(left_proof_str)
        
        right_proof_str = main_tokens[2] if len(main_tokens) > 2 else "null"
        if right_proof_str != "null":
            tree_obj.right_proof = parse_tree(right_proof_str)
        
        gamma_str = sequents[0]
        delta_str = sequents[1]
        
        parse_sequent(gamma_str, tree_obj.gamma)
        parse_sequent(delta_str, tree_obj.delta)
        
        return tree_obj
    else:
        return None

def get_latex_str(t):
    s = '\\[\n\\begin{prooftree}\n'
    s += t.to_latex()[3:-2]
    s += '\\end{prooftree}\n\\]'
    s = [line for line in s.splitlines() if line.strip()]
    s = "\n".join(s)
    return s


if __name__ == '__main__':
    # Stringa di esempio
    s = 'tree(cs, tree(condR, tree(id, tree(assioma, null, null, [[[Atom("437509"), Atom("273413")]], [[Atom("428933"), Atom("273413"), Atom("437509")]], []], [[[Atom("437509"), Atom("273413")], [Atom("428933"), Atom("273413")]], [], [[Atom("428933"), Functor(8708365,2,a,a)]]]), null, [[], [[Atom("428933"), Atom("273413"), Atom("437509")]], []], [[[Atom("437509"), Atom("273413")], [Atom("428933"), Atom("273413")]], [], [[Atom("428933"), Functor(8708365,2,a,a)]]]), null, [[], [], []], [[[Atom("428933"), Atom("273413")]], [], [[Atom("428933"), Functor(8708365,2,a,a)]]]), tree(condR, tree(id, tree(assioma, null, null, [[[Atom("437509"), Atom("273413")]], [[Atom("428933"), Atom("273413"), Atom("437509")]], []], [[[Atom("437509"), Atom("273413")], [Atom("428933"), Atom("273413")]], [], [[Atom("428933"), Functor(8708365,2,a,a)]]]), null, [[], [[Atom("428933"), Atom("273413"), Atom("437509")]], []], [[[Atom("437509"), Atom("273413")], [Atom("428933"), Atom("273413")]], [], [[Atom("428933"), Functor(8708365,2,a,a)]]]), null, [[], [], []], [[[Atom("428933"), Atom("273413")]], [], [[Atom("428933"), Functor(8708365,2,a,a)]]]), [[], [], []], [[], [], [[Atom("428933"), Functor(8708365,2,a,a)]]])'
    s = clean_string(s)

    parsed_tree = parse_tree(s)
    #print(parsed_tree.__repr__())
    print(get_latex_str(parsed_tree))